"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import inspect
import bisect

import tagexpressions

from radish.errors import HookExecError


class HookImpl:
    """Represent a single Hook Implementation"""

    __slots__ = ["what", "when", "func", "on_tags", "order", "is_formatter", "always"]

    def __init__(
        self, what, when, func, on_tags, order, is_formatter=False, always=False
    ):
        self.what = what
        self.when = when
        self.func = func
        self.on_tags = on_tags
        self.order = order
        self.is_formatter = is_formatter
        self.always = always

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self) -> str:
        return "<HookImpl @{}.{} for tags {} with order {}>".format(
            self.when, self.what, self.on_tags, self.order
        )

    def __hash__(self):
        return hash(
            (
                self.what,
                self.when,
                self.func,
                self.on_tags,
                self.order,
                self.is_formatter,
                self.always,
            )
        )

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return (
            self.what == other.what
            and self.when == other.when  # noqa
            and self.func == other.func  # noqa
            and self.on_tags == other.on_tags  # noqa
            and self.order == other.order  # noqa
            and self.is_formatter == other.is_formatter  # noqa
            and self.always == other.always  # noqa
        )

    def __lt__(self, other):
        return self.order < other.order

    def __le__(self, other):
        return self.order <= other.order

    def __gt__(self, other):
        return self.order > other.order

    def __ge__(self, other):
        return self.order >= other.order


class GeneratorHookImpl:
    """Specialized Hook Implementation for Generator Hooks

    A Generator Hook uses a yield statement to separate
    the `before` and `after` part of a Hook.
    """

    def __init__(self, func):
        self.func = func
        self.generator = None

    def __call__(self, *args, **kwargs):
        if self.generator is None:
            # hook is called the first time, thus we create it and
            # consume the `before` part of it
            self.generator = self.func(*args, **kwargs)
            return next(self.generator)
        else:
            # the hook is called the second time, thus we
            # consume the `after` part of it and expect it to be exhausted
            try:
                return next(self.generator)
            except StopIteration:
                pass  # raised when the generator is exhausted


class HookRegistry:
    """The ``HookRegistry`` keeps track of all declared ``HookImpl``s"""

    DEFAULT_HOOK_ORDER = 100

    GENERATOR_HOOK_NAMES = {
        "for_all",
        "each_feature",
        "each_rule",
        "each_scenario",
        "each_step",
    }

    def __init__(self):
        #: Holds a set of all possible Hook combinations
        self._hooks = {
            "before": {
                "all": [],
                "each_feature": [],
                "each_rule": [],
                "each_scenario": [],
                "each_step": [],
            },
            "after": {
                "all": [],
                "each_feature": [],
                "each_rule": [],
                "each_scenario": [],
                "each_step": [],
            },
        }

    def register(
        self, what, when, func, on_tags, order, is_formatter=False, always=False
    ):
        """Register the given Hook for later execution"""
        if inspect.isgeneratorfunction(func):
            # the registered function is a generator hook
            generator_hook = GeneratorHookImpl(func)
            before_hook_impl = HookImpl(
                what, "before", generator_hook, on_tags, order, is_formatter, always
            )
            after_hook_impl = HookImpl(
                what, "after", generator_hook, on_tags, order * -1, is_formatter, always
            )

            if (
                before_hook_impl in self._hooks["before"][what]
                and after_hook_impl in self._hooks["after"][what]
            ):
                # NOTE: allow a Hook Implementation to be registered multiple times.
                #       This can happend when one hook module imports another in the same
                #       RADISH_BASEDIR.
                return

            # insert the HookImpl in the order given by ``order``.
            bisect.insort_right(self._hooks["before"][what], before_hook_impl)
            bisect.insort_right(self._hooks["after"][what], after_hook_impl)
        else:
            # we have regular hook
            hook_impl = HookImpl(what, when, func, on_tags, order, is_formatter, always)

            if hook_impl in self._hooks[when][what]:
                # NOTE: allow a Hook Implementation to be registered multiple times.
                #       This can happend when one hook module imports another in the same
                #       RADISH_BASEDIR.
                return

            # insert the HookImpl in the order given by ``order``.
            bisect.insort_right(self._hooks[when][what], hook_impl)

    def create_hook_decorators(self, context=None):
        """Create Hook decorators for all hook combinations

        The created Hook decorators are injected into the given ``dict``-like ``context`` object.
        If the given ``context`` is ``None`` the Hooks will be injected into ``globals()``.
        """
        if context is None:
            context = globals()

        created_decorator_names = []
        for when, whats in self._hooks.items():

            class HookProvider:
                """whuat"""

                def __init__(self, when):
                    self.when = when

            when_object = HookProvider(when)

            for what in whats.keys():

                def __create_decorator(what, when):
                    def __decorator(
                        on_tags=None,
                        order=self.DEFAULT_HOOK_ORDER,
                        is_formatter=False,
                        always=False,
                    ):
                        if on_tags is None:
                            on_tags = []

                        def __wrapper(func):
                            self.register(
                                what, when, func, on_tags, order, is_formatter, always
                            )
                            return func

                        return __wrapper

                    __decorator.__doc__ = """Decorator to register a hook function

                    A hook function registered with this decorator will be run {when} {what}.

                    Args:
                        on_tags (list): a list of :class:`Tag` names for which this hook will be run
                        order (int): a number which is used to order the registered hooks when running them
                        is_formatter (bool): flag to indicate that the hook is a formatter.
                                             Formatter hooks are run even if ``on_tags`` do not match
                        always (bool): flag to indicate that the hook should always be run.
                                       Only enable this ``True`` if the hook doesn't depend on the Feature File
                                       you will be running.
                    """.format(  # noqa
                        when=when, what=what
                    )

                    return __decorator

                setattr(when_object, what, __create_decorator(what, when))

            context[when] = when_object
            created_decorator_names.append(when)
        return created_decorator_names

    def create_generator_hook_decorators(self, context=None):
        """Create Generator Hook decorators for models

        The created Hook decorators are injected into the given ``dict``-like ``context`` object.
        If the given ``context`` is ``None`` the Hooks will be injected into ``globals()``.
        """
        if context is None:
            context = globals()

        created_decorator_names = []
        for what in self.GENERATOR_HOOK_NAMES:

            def __create_decorator(what):
                def __decorator(
                    on_tags=None,
                    order=self.DEFAULT_HOOK_ORDER,
                    is_formatter=False,
                    always=False,
                ):
                    if on_tags is None:
                        on_tags = []

                    def __wrapper(func):
                        self.register(
                            what, None, func, on_tags, order, is_formatter, always
                        )
                        return func

                    return __wrapper

                __decorator.__doc__ = """Decorator to register a generator hook function

                A generator hook function registered with this decorator will be run
                twice for {what}, once for the `before` part before the yield statement and once
                after the yield statement for the `after` part.

                .. code-block:: python

                   @{what}
                   def do_{what}(model):
                       # do some setup
                       setup()
                       yield
                       # do some teardown
                       teardown()

                Args:
                    on_tags (list): a list of :class:`Tag` names for which this hook will be run
                    order (int): a number which is used to order the registered hooks when running them
                    is_formatter (bool): flag to indicate that the hook is a formatter.
                                         Formatter hooks are run even if ``on_tags`` do not match
                    always (bool): flag to indicate that the hook should always be run.
                                   Only enable this ``True`` if the hook doesn't depend on the Feature File
                                   you will be running.
                """.format(  # noqa
                    what=what
                )

                return __decorator

            context[what] = __create_decorator(what)
            created_decorator_names.append(what)
        return created_decorator_names

    def call(self, what, when, only_formatters, tagged_model, *args, **kwargs):
        """Calls a registered Hook"""
        hooks = self._hooks[when][what]
        for hook_impl in hooks:
            if not hook_impl.always:
                if only_formatters and not hook_impl.is_formatter:
                    continue

            #: holds a flag whether or not the Hook actually has to be called
            #  which is depenend on the `on_tags` setting of the HookImpl.
            call_hook = True

            if hook_impl.on_tags:
                tag_expression = tagexpressions.parse(" or ".join(hook_impl.on_tags))
                # get the Tags for models which actually have Tags
                tags = (
                    tagged_model.get_all_tags()
                    if hasattr(tagged_model, "get_all_tags")
                    else []
                )
                call_hook = tag_expression.evaluate([t.name for t in tags])

            if not call_hook:
                continue

            try:
                hook_impl(tagged_model, *args, **kwargs)
            except Exception as exc:
                raise HookExecError(hook_impl, exc) from exc


#: Holds a global instance of the HookRegistry which shall be used
#  by all modules implementing Hooks.
registry = HookRegistry()
__all__ = (
    registry.create_hook_decorators() + registry.create_generator_hook_decorators()
)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import base64

import radish.utils as utils
from radish.errors import RadishError, StepBehaveLikeRecursionError
from radish.models.state import State
from radish.models.stepfailurereport import StepFailureReport
from radish.models.timed import Timed


class Step(Timed):
    """Respresents a single instance of a Gherkin Step"""

    def __init__(
        self,
        step_id: int,
        keyword: str,
        used_keyword: str,
        text: str,
        doc_string,
        data_table,
        path: str,
        line: int,
    ) -> None:
        super().__init__()
        self.id = step_id
        self.keyword = keyword
        self.used_keyword = used_keyword
        self.text = text
        self.doc_string = doc_string
        self.data_table = data_table
        self.path = path
        self.line = line

        #: Holds information about the AST hierarchy where this Step appeared.
        self.feature = None
        self.rule = None
        self.scenario = None

        #: Holds information about the Step Implementation this Step was matched with.
        self.step_impl = None
        self.step_impl_match = None

        #: Holds the behave-like runner
        self._behave_like_runner = None

        #: Holds information about the State of this Step
        self.state = State.UNTESTED
        self.failure_report = None

        #: Holds user-defined embeddings for this Step
        self.embeddings = []

    def __repr__(self) -> str:
        return "<Step: {id} '{keyword} {text}' @ {path}:{line}>".format(
            id=self.id,
            keyword=self.keyword,
            text=self.text,
            path=self.path,
            line=self.line,
        )  # pragma: no cover

    def set_feature(self, feature):
        """Set the Feature for this Step"""
        self.feature = feature  # pragma: no cover

    def set_rule(self, rule):
        """Set the Rule for this Step"""
        self.rule = rule  # pragma: no cover

    def set_scenario(self, scenario):
        """Set the Scenario for this Step"""
        self.scenario = scenario

    @property
    def context(self):
        """Return the Context object where this Step is running in"""
        return self.scenario.context

    def get_all_tags(self):
        """Return all Tags for this Step

        These Tags include the ones inherited from the Scenario and Feature.
        """
        return self.scenario.get_all_tags()

    def assign_implementation(self, step_impl, match):
        """Assign a matched Step Implementation to this Step"""
        self.step_impl = step_impl
        self.step_impl_match = match

    def __validate_if_runnable(self):
        """Validate if this Step is able to run

        If it's not able to run an Exception is raised.
        """
        if not self.step_impl or not self.step_impl_match:
            raise RadishError(
                "Unable to run Step '{} {}' because it has "
                "no Step Implementation assigned to it".format(self.keyword, self.text)
            )

        if self.state is not State.UNTESTED:
            raise RadishError(
                "Unable to run Step '{} {}' again. A Step can only be run exactly once.".format(
                    self.keyword, self.text
                )
            )

    def with_behave_like_runner(func):
        def __wrapper(self, behave_like_runner, *args, **kwargs):
            self._behave_like_runner = behave_like_runner
            try:
                return func(self, *args, **kwargs)
            finally:
                self._behave_like_runner = None

        return __wrapper

    @with_behave_like_runner
    def run(self):
        """Run this Step

        The Step will only run if a ``StepImpl`` was assigned using ``assign_implementation``.
        """
        self.__validate_if_runnable()
        args, kwargs = self.step_impl_match.evaluate()

        self.state = State.RUNNING
        try:
            if kwargs:
                self.step_impl.func(self, **kwargs)
            else:
                self.step_impl.func(self, *args)
        except Exception as exc:
            self.fail(exc)
        else:
            if self.state is State.RUNNING:
                self.state = State.PASSED
        return self.state

    @with_behave_like_runner
    def debug(self):
        """Run this Step in a debugger"""
        self.__validate_if_runnable()
        args, kwargs = self.step_impl_match.evaluate()

        pdb = utils.get_debugger()

        self.state = State.RUNNING
        try:
            pdb.runcall(self.step_impl.func, self, *args, **kwargs)
        except Exception as exc:
            self.fail(exc)
        else:
            if self.state is State.RUNNING:
                self.state = State.PASSED
        return self.state

    def behave_like(self, step_line):
        """Run this Step as if it would be the one given in ``step_line``

        This function requires ``self._behave_like_runner`` to be set.
        """
        if self._behave_like_runner is None:
            raise RadishError(
                "This Step is unable to use the `behave_like`-Feature because no runner is provided"
            )

        state, step = self._behave_like_runner(step_line)
        if state is State.FAILED:
            try:
                raise step.failure_report.exception
            except RecursionError:
                raise StepBehaveLikeRecursionError()

    def fail(self, exception):
        """Let this Step fail with the given exception"""
        self.state = State.FAILED
        self.failure_report = StepFailureReport(exception)

    def skip(self):
        """Skip this Step

        This method is supposed to be called from Step Implementations.
        """
        if self.state is not State.RUNNING:
            raise RadishError("Steps can only be skipped when they run")
        self.state = State.SKIPPED

    def pending(self):
        """Mark this Step as pending

        All pending Steps will be reminded about in the Runs summary.

        This method is supposed to be called from Step Implementations.
        """
        if self.state is not State.RUNNING:
            raise RadishError("Steps can only be marked as pending when they run")
        self.state = State.PENDING

    def embed(self, data, mime_type="text/plain", encode_data_to_base64=True):
        """Embed data into Step
            - step embedded data can be used for cucumber json reports
            - allow to attach text, html or images

        This method is supposed to be called from Step Implementations.

        :param data: data attached to report
            data needs to be encoded in base64 for proper handling by tests reporting tools
            if not encoded in base64 encoded_data_to_base64 needs to be True (default value)
        :param mime_type: image/png, image/bmp, text/plain, text/html
            - mime types with special support by Cucumber reporting
            - other mime types will be handled with default handling
        :param encode_data_to_base64: encode data to base64 if True
        """
        if encode_data_to_base64:
            data = base64.standard_b64encode(data.encode()).decode()
        self.embeddings.append({"data": data, "mime_type": mime_type})



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from tagexpressions import parse

from radish.errors import RadishError


class Config:
    """Represents the configuration of a radish test run"""

    def __init__(self, command_line_config):
        for opt_key, opt_value in command_line_config.items():
            setattr(self, opt_key, opt_value)

        # make some configuration options easily accessible
        if hasattr(self, "tags") and self.tags:
            try:
                self.tag_expression = parse(self.tags)
            except Exception as exc:
                raise RadishError(
                    "The given Tag Expression '{}' has Syntax Errors. "
                    "Please consult https://github.com/timofurrer/tag-expressions "
                    "for detailed information about the Tag Expression Syntax".format(
                        self.tags
                    )
                ) from exc
        else:
            self.tag_expression = None



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import time
from datetime import timedelta


class Timed:
    """Base-Class with can be used to record a start- and end time on an instance"""

    def __init__(self):
        self.starttime = None
        self.endtime = None

    def record_starttime(self):
        self.starttime = time.time()

    def record_endtime(self):
        self.endtime = time.time()

    def duration(self):
        """Get the current duration.

        If no ``self.endtime`` is set the duration
        is until the current time.
        """
        if self.starttime is None:
            return timedelta(0)

        if self.endtime is None:
            return timedelta(seconds=time.time() - self.starttime)

        return timedelta(seconds=self.endtime - self.starttime)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

__description__ = "Behaviour-Driven-Development tool for python"
__license__ = "MIT"
__version__ = "1.0.0a3"
__author__ = "Timo Furrer"
__author_email__ = "tuxtimo@gmail.com"
__url__ = "http://radish-bdd.io"
__download_url__ = "https://github.com/radish-bdd/radish"
__bugtrack_url__ = "https://github.com/radish-bdd/radish/issues"


# Expose useful objects on radish package level
from radish.errors import RadishError  # noqa
from radish.hookregistry import (  # noqa
    after,
    before,
    each_feature,
    each_rule,
    each_scenario,
    each_step,
    for_all,
)
from radish.models import (  # noqa
    Background,
    ConstantTag,
    DefaultRule,
    Feature,
    PreconditionTag,
    Rule,
    Scenario,
    ScenarioLoop,
    ScenarioOutline,
    Step,
    Tag,
)
from radish.parser import FeatureFileParser  # noqa
from radish.parsetyperegistry import (  # noqa
    TypeBuilder,
    custom_type,
    register_custom_type,
)
from radish.stepregistry import given, step, then, when  # noqa
from radish.terrain import world  # noqa



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

from radish.errors import RadishError
from radish.models.state import State
from radish.models.step import Step
from radish.models.stepfailurereport import StepFailureReport


def test_step_is_initialized_without_a_step_impl():
    """A Step is initialized without a Step Implementation assigned to it"""
    # when
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)

    # then
    assert step.step_impl is None
    assert step.step_impl_match is None


def test_step_is_intialized_in_untested_state():
    """A Step is initialized in the UNTESTED State"""
    # when
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)

    # then
    assert step.state is State.UNTESTED


def test_step_is_intialized_without_a_failure_report():
    """A Step is initialized without a Failure report"""
    # when
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)

    # then
    assert step.failure_report is None


def test_step_is_intialized_without_any_embeddings():
    """A Step is initialized without any embeddings"""
    # when
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)

    # then
    assert step.embeddings == []


def test_step_context_returns_the_same_as_scenario_context(mocker):
    """A Steps context returns the Scenarios context it belongs to"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    scenario_mock = mocker.MagicMock(name="Scenario")
    scenario_mock.context = mocker.MagicMock(name="Scenario Context")
    step.set_scenario(scenario_mock)

    # then
    assert step.context is scenario_mock.context


def test_step_can_assign_a_step_impl(mocker):
    """A Step can be assigned a Step Implementation"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step_impl_mock = mocker.MagicMock(name="Step Impl")
    step_impl_match_mock = mocker.MagicMock(name="Step Impl Match")

    # when
    step.assign_implementation(step_impl_mock, step_impl_match_mock)

    # then
    assert step.step_impl is step_impl_mock
    assert step.step_impl_match is step_impl_match_mock


def test_step_fail_to_run_if_no_step_impl():
    """A Step should fail to run if it has no Step Implementation assigned to it"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)

    # then
    with pytest.raises(RadishError):
        # when
        step.run(None)


def test_step_fail_to_run_if_already_run(mocker):
    """A Step should fail to run if it was already run / has another state then UNTESTED"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step_impl_mock = mocker.MagicMock(name="Step Impl")
    step_impl_match_mock = mocker.MagicMock(name="Step Impl Match")
    step.assign_implementation(step_impl_mock, step_impl_match_mock)
    step.state = State.PASSED

    # then
    with pytest.raises(RadishError):
        # when
        step.run(None)


def test_step_should_evaluate_its_matched_step_impl_arguments(mocker):
    """A Step should evlauate the arguments of its matched Step Implementation"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step_impl_mock = mocker.MagicMock(name="Step Impl")
    step_impl_match_mock = mocker.MagicMock(name="Step Impl Match")
    step_impl_match_mock.evaluate.return_value = ([], {})
    step.assign_implementation(step_impl_mock, step_impl_match_mock)

    # when
    step.run(None)

    # then
    step_impl_match_mock.evaluate.assert_called_once_with()


def test_step_should_set_state_to_running_before_running_step_impl(mocker):
    """A Step should set its State to RUNNING before it runs the Step Implementation function"""
    # given
    class WrapperForMockerSpy:
        def step_func(self, step):
            assert step.state is State.RUNNING

    w = WrapperForMockerSpy()
    mocker.spy(w, "step_func")

    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step_impl_mock = mocker.MagicMock(name="Step Impl")
    step_impl_mock.func = w.step_func
    step_impl_match_mock = mocker.MagicMock(name="Step Impl Match")
    step_impl_match_mock.evaluate.return_value = ([], {})
    step.assign_implementation(step_impl_mock, step_impl_match_mock)

    # when
    step.run(None)

    # then
    w.step_func.assert_called_once_with(step)


def test_step_should_pass_evaluated_kwargs_to_step_impl_func(mocker):
    """
    A Step should pass the evaluated kwargs from the Step Implementation match function
    to the Step Implementation function.
    """
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step_impl_mock = mocker.MagicMock(name="Step Impl")
    step_impl_match_mock = mocker.MagicMock(name="Step Impl Match")
    step_impl_match_mock.evaluate.return_value = ([], {"foo": "bar", "bla": "meh"})
    step.assign_implementation(step_impl_mock, step_impl_match_mock)

    # when
    step.run(None)

    # then
    step_impl_mock.func.assert_called_once_with(step, foo="bar", bla="meh")


def test_step_should_pass_evaluated_args_to_step_impl_func(mocker):
    """
    A Step should pass the evaluated args from the Step Implementation match function
    to the Step Implementation function.
    """
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step_impl_mock = mocker.MagicMock(name="Step Impl")
    step_impl_match_mock = mocker.MagicMock(name="Step Impl Match")
    step_impl_match_mock.evaluate.return_value = (["foo", "bar"], {})
    step.assign_implementation(step_impl_mock, step_impl_match_mock)

    # when
    step.run(None)

    # then
    step_impl_mock.func.assert_called_once_with(step, "foo", "bar")


def test_step_fail_if_step_impl_func_raises(mocker):
    """A Step should fail if the Step Implementation function raised an Exception"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step.fail = mocker.MagicMock(name="Step fail function")
    step_impl_mock = mocker.MagicMock(name="Step Impl")
    exception = Exception("buuh!")
    step_impl_mock.func.side_effect = exception
    step_impl_match_mock = mocker.MagicMock(name="Step Impl Match")
    step_impl_match_mock.evaluate.return_value = ([], {})
    step.assign_implementation(step_impl_mock, step_impl_match_mock)

    # when
    step.run(None)

    # then
    step.fail.assert_called_once_with(exception)


def test_step_should_change_state_to_passed_if_step_impl_func_not_raised(mocker):
    """
    A Step should change its State to PASSED if the ran
    Step Implementation function did not raise any Exception
    """
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step_impl_mock = mocker.MagicMock(name="Step Impl")
    step_impl_match_mock = mocker.MagicMock(name="Step Impl Match")
    step_impl_match_mock.evaluate.return_value = ([], {})
    step.assign_implementation(step_impl_mock, step_impl_match_mock)

    # when
    step.run(None)

    # then
    assert step.state is State.PASSED


def test_step_should_fail_with_failed_state_and_report():
    """When a Step is failed it should change its State to FAILED and create a report"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)

    # when
    try:
        raise Exception("buuh!")
    except Exception as exc:
        step.fail(exc)

    # then
    assert step.state is State.FAILED
    assert isinstance(step.failure_report, StepFailureReport)


def test_step_should_be_able_to_skip_while_running():
    """A Step should be able to be skipped while it's running"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step.state = State.RUNNING

    # when
    step.skip()

    # then
    assert step.state is State.SKIPPED


def test_step_should_not_be_able_to_skip_when_not_running():
    """A Step should not be able to be skipped when it's not running"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)

    # then
    with pytest.raises(RadishError):
        # when
        step.skip()

    assert step.state is not State.SKIPPED


def test_step_should_be_able_to_mark_pending_while_running():
    """A Step should be able to be marked pending while it's running"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step.state = State.RUNNING

    # when
    step.pending()

    # then
    assert step.state is State.PENDING


def test_step_should_not_be_able_to_mark_pending_when_not_running():
    """A Step should not be able to be marked pending when it's not running"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)

    # then
    with pytest.raises(RadishError):
        # when
        step.pending()

    assert step.state is not State.PENDING


def test_step_should_embed_data_without_encoding():
    """A Step should be able to embed data without an encoding"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)

    # when
    step.embed("data", encode_data_to_base64=False)

    # then
    assert step.embeddings == [{"data": "data", "mime_type": "text/plain"}]


def test_step_should_embed_data_with_base64_encoding():
    """A Step should be able to embed data base64 encoded"""
    # given
    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    data_b64_encoded = "ZGF0YQ=="  # == "data"

    # when
    step.embed("data", encode_data_to_base64=True)

    # then
    assert step.embeddings == [{"data": data_b64_encoded, "mime_type": "text/plain"}]


def test_step_should_not_set_passed_state_if_state_changed_during_run(mocker):
    """A Step should not set its State to PASSED if the state was changed during the run"""
    # given
    def step_change_state(step):
        step.skip()

    step = Step(1, "keyword", "used_keyword", "text", None, None, None, None)
    step_impl_mock = mocker.MagicMock(name="Step Impl")
    step_impl_mock.func = step_change_state
    step_impl_match_mock = mocker.MagicMock(name="Step Impl Match")
    step_impl_match_mock.evaluate.return_value = ([], {})
    step.assign_implementation(step_impl_mock, step_impl_match_mock)

    # when
    step.run(None)

    # then
    assert step.state is State.SKIPPED



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

from radish.parsetyperegistry import ParseTypeRegistry
from radish.errors import RadishError


def test_parsetyperegistry_should_initialize_without_any_registered_types():
    """The ParseTypeRegistry should be initialized wihtout any pre-registered Custom Parse Types"""
    # when
    registry = ParseTypeRegistry()

    # then
    assert registry.types == {}


def test_parsetyperegistry_should_register_funcs_as_custom_parse_types():
    """The ParseTypeRegistry should register a function with a given name as a Custom Parse Type"""
    # given
    registry = ParseTypeRegistry()
    custom_parse_type_func = lambda x: x  # noqa

    # when
    registry.register("name", custom_parse_type_func, "pattern")

    # then
    assert registry.types == {"name": custom_parse_type_func}


def test_parsetyperegistry_should_only_allow_unique_custom_parse_type_names():
    """
    The ParseTypeRegistry should only allow to register new
    Custom Parse Types with unique names
    """
    # given
    registry = ParseTypeRegistry()
    custom_parse_type_func = lambda x: x  # noqa
    registry.register("name", custom_parse_type_func, "pattern")

    # then
    with pytest.raises(RadishError):
        # when
        registry.register("name", custom_parse_type_func, "pattern")


def test_parsetyperegistry_should_assign_pattern_to_func_during_register():
    """
    The ParseTypeRegistry should assign the Custom Type Parse pattern
    to the function which is registered as type handler.
    """
    # given
    registry = ParseTypeRegistry()
    custom_parse_type_func = lambda x: x  # noqa

    # when
    registry.register("name", custom_parse_type_func, "pattern")

    # then
    assert custom_parse_type_func.pattern == "pattern"


def test_parsetyperegistry_should_create_custom_type_decorator_into_context():
    """
    The ParseTypeRegistry should only allow to register new
    Custom Parse Types with unique names
    """
    # given
    registry = ParseTypeRegistry()
    context = {}

    # when
    created_decorator_name = registry.create_decorator(context)

    # then
    assert created_decorator_name == "custom_type"
    assert "custom_type" in context
    assert callable(context["custom_type"])


def test_parsetyperegistry_should_register_custom_parse_type_via_decorator():
    """
    The ParseTypeRegistry should only to register a Custom Parse Type
    with the created @custom_type decorator.
    """
    # given
    registry = ParseTypeRegistry()
    context = {}
    registry.create_decorator(context)

    decorator = context["custom_type"]

    # when
    @decorator("name", "pattern")
    def my_custom_type(text):
        ...

    # then
    assert registry.types == {"name": my_custom_type}



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from enum import IntEnum


class State(IntEnum):
    """Represent the State of a Model

    A Model can be one of:

    * :class:`Step`
    * :class:`Scenario`
    * :class:`Feature`
    """

    PASSED = 0  #: State which is set for a :class:`Step` if it ran successfully
    UNTESTED = 1  #: Default State for all :class:`Step`
    SKIPPED = 2  #: State which is set for a :class:`Step` when it's skipped using :func:`Step.skip`
    PENDING = 3  #: State which is set for a :class:`Step` when it's marked pending using :func:`Step.pending`  # noqa
    FAILED = 4  #: State which is set for a :class:`Step` when it failed to run it
    RUNNING = 5  #: State which is set while a :class:`Step` is running

    @staticmethod
    def report_state(states):
        """Report the single most appropriate State out of a list of States"""
        try:
            return sorted(states, reverse=True)[0]
        except IndexError:
            # states generator did not yield any States
            return State.UNTESTED



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import logging

# configure the radish command line logger which is used for debugging
logger = logging.getLogger("radish")
logging.basicConfig(
    level=logging.CRITICAL, format="%(asctime)s - %(name)s [%(levelname)s]: %(message)s"
)


def enable_radish_debug_mode_click_hook(ctx, param, enabled):
    """Enable radish for debugging"""
    if enabled:
        logger.setLevel(logging.DEBUG)
        logger.debug("Enabled debug mode")
    else:
        logger.setLevel(logging.ERROR)


__all__ = ["logger", "enable_radish_debug_mode_click_hook"]



"""
radish
~~~~~~

the root from red to green.  BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.parser.core import FeatureFileParser  # noqa



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from unittest.mock import MagicMock

import pytest
from tagexpressions import parse

from radish.models import ConstantTag, Scenario, Tag
from radish.models.state import State


def test_scenario_should_set_feature_to_steps_when_feature_set(mocker):
    """A Scenario should forward the Feature to each Step"""
    # given
    first_step = mocker.MagicMock(name="First Step")
    second_step = mocker.MagicMock(name="Second Step")
    scenario = Scenario(
        1, "Scenario", "My Scenario", [], None, None, [first_step, second_step]
    )
    feature_mock = mocker.MagicMock(name="Feature")

    # when
    scenario.set_feature(feature_mock)

    # then
    assert scenario.feature is feature_mock
    first_step.set_feature.assert_called_once_with(feature_mock)
    second_step.set_feature.assert_called_once_with(feature_mock)


def test_scenario_should_set_rule_to_steps_when_rule_set(mocker):
    """A Scenario should forward the Rule to each Step"""
    # given
    first_step = mocker.MagicMock(name="First Step")
    second_step = mocker.MagicMock(name="Second Step")
    scenario = Scenario(
        1, "Scenario", "My Scenario", [], None, None, [first_step, second_step]
    )
    rule_mock = mocker.MagicMock(name="Rule")

    # when
    scenario.set_rule(rule_mock)

    # then
    assert scenario.rule is rule_mock
    first_step.set_rule.assert_called_once_with(rule_mock)
    second_step.set_rule.assert_called_once_with(rule_mock)


def test_scenario_should_copy_background_when_set(mocker):
    """A Scenario should make a deepcopy when a Background is set"""
    # given
    scenario = Scenario(1, "Scenario", "My Scenario", [], None, None, [])
    background_mock = mocker.MagicMock(name="Background")
    background_mock.short_description = "My Background"

    # when
    scenario.set_background(background_mock)

    # then
    assert scenario.background.short_description == background_mock.short_description
    assert scenario.background is not background_mock


def test_scenario_should_set_itself_as_scenario_for_a_set_background(mocker):
    """A Scenario should set itself as Scenario for a set Background"""
    # given
    scenario = Scenario(1, "Scenario", "My Scenario", [], None, None, [])
    background_mock = mocker.MagicMock(name="Background")

    # when
    scenario.set_background(background_mock)

    # then
    scenario.background.set_scenario.assert_called_once_with(scenario)


def test_scenario_should_set_its_rule_as_rule_for_a_set_background(mocker):
    """A Scenario should set its Rule as Rule for a set Background"""
    # given
    scenario = Scenario(1, "Scenario", "My Scenario", [], None, None, [])
    scenario.rule = mocker.MagicMock(name="Rule")
    background_mock = mocker.MagicMock(name="Background")

    # when
    scenario.set_background(background_mock)

    # then
    scenario.background.set_rule.assert_called_once_with(scenario.rule)


def test_scenario_should_ignore_to_set_none_for_a_background(mocker):
    """A Scenario should ignore to set None for a Background"""
    # given
    scenario = Scenario(1, "Scenario", "My Scenario", [], None, None, [])
    background_mock = mocker.MagicMock(name="Background")
    scenario.set_background(background_mock)

    # when
    scenario.set_background(None)

    # then
    assert scenario.background is not None


@pytest.mark.parametrize(
    "given_background_state, given_steps, expected_state",
    [
        (State.PASSED, [], State.PASSED),
        (
            State.PASSED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
            ],
            State.PASSED,
        ),
        (
            State.PASSED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.FAILED),
                MagicMock(state=State.UNTESTED),
            ],
            State.FAILED,
        ),
        (
            State.PASSED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.UNTESTED),
                MagicMock(state=State.UNTESTED),
            ],
            State.UNTESTED,
        ),
        (
            State.PASSED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.SKIPPED),
                MagicMock(state=State.UNTESTED),
            ],
            State.SKIPPED,
        ),
        (
            State.PASSED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.SKIPPED),
                MagicMock(state=State.PENDING),
            ],
            State.PENDING,
        ),
        (
            State.PASSED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.SKIPPED),
                MagicMock(state=State.FAILED),
            ],
            State.FAILED,
        ),
        (
            State.PASSED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.UNTESTED),
                MagicMock(state=State.FAILED),
            ],
            State.FAILED,
        ),
        (
            State.UNTESTED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
            ],
            State.UNTESTED,
        ),
        (
            State.SKIPPED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
            ],
            State.SKIPPED,
        ),
        (
            State.PENDING,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
            ],
            State.PENDING,
        ),
        (
            State.FAILED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
                MagicMock(state=State.PASSED),
            ],
            State.FAILED,
        ),
        (
            State.SKIPPED,
            [
                MagicMock(state=State.PASSED),
                MagicMock(state=State.FAILED),
                MagicMock(state=State.PASSED),
            ],
            State.FAILED,
        ),
    ],
    ids=[
        "Background PASSED, No Steps -> Scenario PASSED",
        "Background PASSED, Steps all PASSED -> Scenario PASSED",
        "Background PASSED, Steps [PASSED, FAILED, UNTESTED] -> Scenario FAILED",
        "Background PASSED, Steps [PASSED, UNTESTED, UNTESTED] -> Scenario UNTESTED",
        "Background PASSED, Steps [PASSED, SKIPPED, UNTESTED] -> Scenario SKIPPED",
        "Background PASSED, Steps [PASSED, SKIPPED, PENDING] -> Scenario PENDING",
        "Background PASSED, Steps [PASSED, SKIPPED, FAILED] -> Scenario FAILED",
        "Background PASSED, Steps [PASSED, UNTESTED, FAILED] -> Scenario FAILED",
        "Background UNTESTED, Steps [PASSED, PASSED, PASSED] -> Scenario UNTESTED",
        "Background SKIPPED, Steps [PASSED, PASSED, PASSED] -> Scenario SKIPPED",
        "Background PENDING, Steps [PASSED, PASSED, PASSED] -> Scenario PENDING",
        "Background FAILED, Steps [PASSED, PASSED, PASSED] -> Scenario FAILED",
        "Background SKIPPED, Steps [PASSED, FAILED, PASSED] -> Scenario FAILED",
    ],
)
def test_scenario_should_return_correct_state(
    given_background_state, given_steps, expected_state, mocker
):
    """
    A Scneario should return the correct State depending on its
    Background State and its own Steps State.
    """
    # given
    scenario = Scenario(1, "Scenario", "My Scenario", [], None, None, given_steps)
    background_mock = mocker.MagicMock(name="Background")
    background_mock.state = given_background_state
    background_mock.steps = [mocker.MagicMock(state=given_background_state)]
    scenario.set_background(background_mock)

    # when
    actual_state = scenario.state

    # then
    assert actual_state == expected_state


def test_scenario_without_steps_should_return_untested_state(mocker):
    """
    A Scneario without any Steps should return the State UNTESTED
    """
    # given
    scenario = Scenario(1, "Scenario", "My Scenario", [], None, None, [])
    scenario.background = None

    # when
    actual_state = scenario.state

    # then
    assert actual_state == State.UNTESTED


def test_scenario_should_ignore_background_state_if_no_background(mocker):
    """A Scenario should ignore the Background for the State if no is assigned"""
    # given
    scenario = Scenario(
        1,
        "Scenario",
        "My Scenario",
        [],
        None,
        None,
        [mocker.MagicMock(state=State.SKIPPED)],
    )

    # when
    actual_state = scenario.state

    # then
    assert actual_state == State.SKIPPED


@pytest.mark.parametrize(
    "tagexpression, scenario_ids, expected_has_to_run",
    [
        (None, [], True),
        (parse("tag-a"), [], True),
        (parse("tag-c"), [], True),
        (parse("tag-X"), [], False),
        (None, [1], True),
        (None, [2], False),
        (parse("tag-a"), [2], False),
        (parse("tag-X"), [1], False),
        (parse("tag-a"), [1], True),
    ],
    ids=[
        "no tagexpression, no scenario_ids => RUN",
        "tagexpression match in Scenario Tags, no scenario_ids => RUN",
        "tagexpression match in Feature Tags, no scenario_ids => RUN",
        "tagexpression no match, no scenario_ids => NO RUN",
        "no tagexpression, scenario_ids match => RUN",
        "no tagexpression, scenario_ids no match => NO RUN",
        "tagexpression match, scenario_ids no match => NO RUN",
        "tagexpression no match, scenario_ids match => NO RUN",
        "tag expression match, scenario_ids match => RUN",
    ],
)
def test_scenario_should_correctly_evaluate_if_it_has_to_be_run(
    mocker, tagexpression, scenario_ids, expected_has_to_run
):
    """Test that a Scenario should correctly evaluate if it has to be run or not"""
    # given
    feature_mock = mocker.MagicMock(tags=[Tag("tag-c", None, None)])
    scenario = Scenario(
        1,
        "Scenario",
        "My Scenario",
        [Tag("tag-a", None, None), Tag("tag-b", None, None)],
        None,
        None,
        [],
    )
    scenario.set_feature(feature_mock)

    # when
    has_to_run = scenario.has_to_run(tagexpression, scenario_ids)

    # then
    assert has_to_run == expected_has_to_run


def test_scenario_should_return_all_constants(mocker):
    """A Scenario should return all Constants from the Tags"""
    # given
    tags = [
        Tag("x", None, None),
        ConstantTag("k1", "v1", None, None),
        Tag("y", None, None),
        ConstantTag("k2", "v2", None, None),
    ]
    scenario = Scenario(1, "Scenario", "My Scenario", tags, None, None, [])
    feature_mock = mocker.MagicMock(name="Feature")
    feature_mock.constants = {}
    scenario.set_feature(feature_mock)

    # when
    actual_constants = scenario.constants

    # then
    assert actual_constants == {"k1": "v1", "k2": "v2"}


def test_scenario_should_inherit_feature_constants(mocker):
    """A Scenario should inherit all Constants from the Feature"""
    # given
    tags = [
        Tag("x", None, None),
        ConstantTag("k1", "v1", None, None),
        Tag("y", None, None),
        ConstantTag("k2", "v2", None, None),
    ]
    feature_mock = mocker.MagicMock(name="Feature")
    feature_mock.constants = {"kf1": "vf1", "kf2": "vf2"}
    scenario = Scenario(1, "Scenario", "My Scenario", tags, None, None, [])
    scenario.set_feature(feature_mock)

    # when
    actual_constants = scenario.constants

    # then
    assert actual_constants == {"k1": "v1", "k2": "v2", "kf1": "vf1", "kf2": "vf2"}



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from unittest import mock

import pytest

from radish.hookregistry import GeneratorHookImpl, HookImpl, HookRegistry
from radish.models import Tag


@pytest.mark.parametrize(
    "hook_1, hook_2, expect_equal",
    [
        (
            HookImpl("what", "when", None, [], 1),
            HookImpl("what", "when", None, [], 1),
            True,
        ),
        (
            HookImpl("what", "when", None, [], 1),
            HookImpl("", "when", None, [], 1),
            False,
        ),
        (
            HookImpl("what", "when", None, [], 1),
            HookImpl("what", "", None, [], 1),
            False,
        ),
        (
            HookImpl("what", "when", None, [], 1),
            HookImpl("what", "when", lambda x: x, [], 1),
            False,
        ),
        (
            HookImpl("what", "when", None, [], 1),
            HookImpl("what", "when", None, [1], 1),
            False,
        ),
        (
            HookImpl("what", "when", None, [], 1),
            HookImpl("what", "when", None, [], 2),
            False,
        ),
    ],
)
def test_hookimpls_can_be_compared_by_equality(hook_1, hook_2, expect_equal):
    """The ``HookImpl``s can be compared for their equality"""
    # when
    are_equal = hook_1 == hook_2

    # then
    assert are_equal == expect_equal


def test_hookimpls_can_be_called():
    # given
    def func(x, y, foo=None):
        return x, y, foo

    hook = HookImpl("what", "when", func, [], 1)

    # when
    x, y, foo = hook(1, 2, foo=3)

    # then
    assert x == 1
    assert y == 2
    assert foo == 3


def test_generatorhookimpls_can_be_called_twice_to_exhaust_generator():
    # given
    def func(x, y, foo=None):
        x += 1
        y += 1
        foo += 1
        yield x, y, foo
        x += 1
        y += 1
        foo += 1
        yield x, y, foo

    hook = GeneratorHookImpl(func)

    # when
    first_x, first_y, first_foo = hook(1, 2, foo=3)
    second_x, second_y, second_foo = hook(1, 2, foo=3)

    # then
    assert first_x == 2
    assert first_y == 3
    assert first_foo == 4
    assert second_x == 3
    assert second_y == 4
    assert second_foo == 5


def test_hookimpls_can_be_sorted_by_the_order():
    """The ``HookImpl``s can be sorted by it's order"""
    # given
    hooks = [
        HookImpl("what", "when", None, [], 1),
        HookImpl("what", "when", None, [], 10),
        HookImpl("what", "when", None, [], 5),
        HookImpl("what", "when", None, [], 2),
        HookImpl("what", "when", None, [], 30),
        HookImpl("what", "when", None, [], 8),
        HookImpl("what", "when", None, [], 7),
    ]

    # when
    sorted_hooks = sorted(hooks)

    # then
    assert sorted_hooks == [
        HookImpl("what", "when", None, [], 1),
        HookImpl("what", "when", None, [], 2),
        HookImpl("what", "when", None, [], 5),
        HookImpl("what", "when", None, [], 7),
        HookImpl("what", "when", None, [], 8),
        HookImpl("what", "when", None, [], 10),
        HookImpl("what", "when", None, [], 30),
    ]


def test_hookregistry_module_should_have_global_registry_instance():
    """The radish.hookregistry module should contain a global HookRegistry instance"""
    # given & when
    from radish.hookregistry import registry

    # then
    assert isinstance(registry, HookRegistry)


def test_hookregistry_module_should_have_global_hook_decorators():
    """The radish.hookregistry module should contain functions for the Hook decorators"""
    # given & when
    from radish.hookregistry import before, after

    # then
    assert callable(before.all)
    assert callable(before.each_feature)
    assert callable(before.each_rule)
    assert callable(before.each_scenario)
    assert callable(before.each_step)
    assert callable(after.all)
    assert callable(after.each_feature)
    assert callable(after.each_rule)
    assert callable(after.each_scenario)
    assert callable(after.each_step)


def test_hookregistry_module_should_have_global_generator_hook_decorators():
    # given & when
    from radish.hookregistry import (
        for_all,
        each_feature,
        each_rule,
        each_scenario,
        each_step,
    )

    # then
    assert callable(for_all)
    assert callable(each_feature)
    assert callable(each_rule)
    assert callable(each_scenario)
    assert callable(each_step)


def test_hookregistry_call_hook_func_if_match(mocker):
    # given
    registry = HookRegistry()

    to_be_called_hook_func = mocker.MagicMock(name="before_each_scenario_func")
    not_to_be_called_hook = mocker.MagicMock(name="after_each_scenario_func")

    registry.register(
        what="each_scenario",
        when="before",
        func=to_be_called_hook_func,
        on_tags=[],
        order=1,
    )
    registry.register(
        what="each_scenario",
        when="after",
        func=not_to_be_called_hook,
        on_tags=[],
        order=1,
    )

    # when
    registry.call(
        "each_scenario", "before", False, mocker.MagicMock(name="TaggedModel")
    )

    # then
    to_be_called_hook_func.assert_has_calls([mock.call(mock.ANY)])
    not_to_be_called_hook.assert_has_calls([])


def test_hookregistry_call_hook_func_if_tag_matched(mocker):
    # given
    registry = HookRegistry()

    to_be_called_hook_func = mocker.MagicMock(name="before_each_scenario_func_tagged")
    not_to_be_called_hook = mocker.MagicMock(name="before_each_scenario_func")

    registry.register(
        what="each_scenario",
        when="before",
        func=to_be_called_hook_func,
        on_tags=["some-tag"],
        order=1,
    )
    registry.register(
        what="each_scenario",
        when="before",
        func=not_to_be_called_hook,
        on_tags=["another-tag"],
        order=1,
    )

    tagged_model = mocker.MagicMock(name="TaggedModel")
    tagged_model.get_all_tags.return_value = [Tag(name="some-tag", path="", line=0)]

    # when
    registry.call("each_scenario", "before", False, tagged_model)

    # then
    to_be_called_hook_func.assert_has_calls([mock.call(mock.ANY)])
    not_to_be_called_hook.assert_has_calls([])



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import re

from parse_type.cfparse import Parser

from radish.parsetyperegistry import registry as parsetype_registry
from radish.models.step import Step
from radish.errors import (
    StepImplementationNotFoundError,
    StepImplementationPatternNotSupported,
)

try:
    re_pattern_type = re.Pattern  # >= Python 3.7
except AttributeError:
    re_pattern_type = re._pattern_type


class RegexStepImplMatcher:
    """Matcher for Regular Expression based Step Implementations"""

    class Match:
        def __init__(self, match):
            self.match = match

        def evaluate(self):
            """Lazy and return evaluate the step group matches"""
            return self.match.groups(), self.match.groupdict()

    def __call__(self, step_text, step_impl):
        match = step_impl.pattern.search(step_text)
        if match:
            return self.Match(match), len(match.group())

        return None, 0


class ParseTypeStepImplMatcher:
    """Matcher for parse-type based Step Implementations"""

    class Match:
        def __init__(self, match):
            self.match = match

        def evaluate(self):
            """Lazy and return evaluate the step group matches"""
            result = self.match.evaluate_result()
            return result.fixed, result.named

    def __call__(self, step_text, step_impl):
        try:
            parser = Parser(step_impl.pattern, parsetype_registry.types)
        except ValueError:
            raise Exception("Cannot create parser")
            # raise StepPatternError(step_impl.pattern, func.__name__, e)

        match = parser.search(step_text, evaluate_result=False)
        if match:
            return self.Match(match), len(match.match.group())

        return None, 0


#: Holds a mapping of available Matchers
matchers = {re_pattern_type: RegexStepImplMatcher(), str: ParseTypeStepImplMatcher()}


def match_step(step: Step, registry):
    """Match the given ``Step`` with a registered Step Implementation from the ``StepRegistry``.

    If no match can be made an error is raised.
    If a match can be made the found ``StepImpl`` is assigned to the ``Step``.
    """
    # get all possible Step Implementation from the registry
    # depending on the Step Keyword.
    step_impls = registry.step_implementations(step.keyword)

    # resolve the Constant Tags for the matching
    # FIXME(TF): it's quite shitty that the matcher has to know about constants ...
    step_text = _resolve_constant_tags_in_step_text(step.text, step.scenario.constants)

    potentional_matches = []
    for step_impl in step_impls:
        try:
            matcher = matchers[type(step_impl.pattern)]
        except KeyError:
            raise StepImplementationPatternNotSupported(step_impl)

        match, match_length = matcher(step_text, step_impl)
        if match:
            if len(step_text) == match_length:
                # if perfect match can be made we return it no
                # matter of the other potentional matches
                step.assign_implementation(step_impl, match)
                return

            distance_to_perfect = abs(len(step_text) - match_length)
            potentional_matches.append(((step_impl, match), distance_to_perfect))

    if potentional_matches:
        # get best match
        best_step_impl, best_match = min(potentional_matches, key=lambda x: x[1])[0]
        step.assign_implementation(best_step_impl, best_match)
        return

    raise StepImplementationNotFoundError(step)


def _resolve_constant_tags_in_step_text(step_text, constants):
    """Resolve all Constants in the given Step Text

    The Constants must have the form of: ``${name}``.

    >>> step_text = "Some Step ${x} with ${y} and Constants"
    >>> constants = {"x": "A", "y": "B"}
    >>> _resolve_constant_tags_in_step_text(step_text, constants)
    'Some Step A with B and Constants'

    If a ``${name}`` is not in the Constants it's not replaced
    and doesn't cause a warning. It's assumed to be part of the Step itself.

    >>> step_text = "Some Step ${x} with ${y} and Constants"
    >>> constants = {"x": "A"}
    >>> _resolve_constant_tags_in_step_text(step_text, constants)
    'Some Step A with ${y} and Constants'
    """
    for key, value in constants.items():
        step_text = step_text.replace("${%s}" % key, value)

    return step_text



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import re
import textwrap

import colorful as cf
import pytest

from radish.formatters.gherkin import (
    write_feature_footer,
    write_feature_header,
    write_rule_header,
    write_scenario_footer,
    write_scenario_header,
    write_step,
    write_step_result,
    write_tagline,
)
from radish.models import (
    Background,
    DefaultRule,
    Feature,
    Rule,
    Scenario,
    State,
    Step,
    Tag,
)


@pytest.fixture(name="disabled_colors", scope="function")
def disable_ansi_colors(world_default_config):
    """Fixture to disable ANSI colors"""
    orig_colormode = cf.colormode
    cf.disable()
    world_default_config.no_ansi = True
    yield
    cf.colormode = orig_colormode


def dedent_feature_file(contents):
    """Dedent the given Feature File contents"""
    dedented = textwrap.dedent(contents)
    # remove first empty line
    return "\n".join(dedented.splitlines()[1:]) + "\n"


def assert_output(capsys, expected_stdout):
    """Assert that the captured stdout matches"""
    actual_stdout = capsys.readouterr().out
    for actual_stdout_line, expected_stdout_line in zip(
        actual_stdout.splitlines(), expected_stdout.splitlines()
    ):
        assert re.match(
            "^" + expected_stdout_line + "$", actual_stdout_line
        ), "{!r} == {!r}".format(expected_stdout_line, actual_stdout_line)


def test_gf_write_tag_after_an_at_sign(disabled_colors, capsys, mocker):
    """Test that the Gherkin Formatter writes a Tag after the @-sign on a single line"""
    # given
    tag = mocker.MagicMock(spec=Tag)
    tag.name = "tag-a"

    # when
    write_tagline(tag)

    # then
    stdout = capsys.readouterr().out
    assert stdout == "@tag-a\n"


def test_gf_write_tag_with_given_identation(disabled_colors, capsys, mocker):
    """Test that the Gherkin Formatter writes a Tag with the given indentation"""
    # given
    tag = mocker.MagicMock(spec=Tag)
    tag.name = "tag-a"
    indentation = " " * 4

    # when
    write_tagline(tag, indentation)

    # then
    stdout = capsys.readouterr().out
    assert stdout == "    @tag-a\n"


def test_gf_write_feature_header_without_tags_without_description_without_background(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly writes a Feature header
    with no Tags, no description and no Background
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.keyword = "Feature"
    feature.short_description = "My Feature"
    feature.description = []
    feature.tags = []
    feature.background = None

    # when
    write_feature_header(feature)

    # then
    stdout = capsys.readouterr().out
    assert stdout == "Feature: My Feature\n"


def test_gf_write_feature_header_with_tags_without_description_without_background(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly writes a Feature header
    with Tags, but no description and no Background
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.keyword = "Feature"
    feature.short_description = "My Feature"
    feature.description = []
    first_tag = mocker.MagicMock(spec=Tag)
    first_tag.name = "tag-a"
    second_tag = mocker.MagicMock(spec=Tag)
    second_tag.name = "tag-b"
    feature.tags = [first_tag, second_tag]
    feature.background = None

    # when
    write_feature_header(feature)

    # then
    stdout = capsys.readouterr().out
    assert stdout == dedent_feature_file(
        """
        @tag-a
        @tag-b
        Feature: My Feature
        """
    )


def test_gf_write_feature_header_without_tags_with_description_without_background(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly writes a Feature header
    without Tags and Background, but description
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.keyword = "Feature"
    feature.short_description = "My Feature"
    feature.description = ["foo", "bar", "bla"]
    feature.tags = []
    feature.background = None

    # when
    write_feature_header(feature)

    # then
    stdout = capsys.readouterr().out
    assert stdout == dedent_feature_file(
        """
        Feature: My Feature
            foo
            bar
            bla

        """
    )


def test_gf_write_feature_header_without_description_with_empty_background_no_short_description(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly writes a Feature header
    without Tags and Description, but with an empty Background with no short description
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.keyword = "Feature"
    feature.short_description = "My Feature"
    feature.description = []
    feature.tags = []
    feature.background = mocker.MagicMock(spec=Background)
    feature.background.keyword = "Background"
    feature.background.short_description = None
    feature.background.steps = []

    # when
    write_feature_header(feature)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            Feature: My Feature
                Background:[ ]

            """
        ),
    )


def test_gf_write_feature_header_with_description_with_empty_background_no_short_description(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly writes a Feature header
    without Tags, but Description and an empty Background with no short description
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.keyword = "Feature"
    feature.short_description = "My Feature"
    feature.description = ["foo", "bar", "bla"]
    feature.tags = []
    feature.background = mocker.MagicMock(spec=Background)
    feature.background.keyword = "Background"
    feature.background.short_description = None
    feature.background.steps = []

    # when
    write_feature_header(feature)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            Feature: My Feature
                foo
                bar
                bla

                Background:[ ]

            """
        ),
    )


def test_gf_write_feature_header_empty_background_with_short_description(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly writes a Feature header
    without Tags and Description but an empty Background with short description
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.keyword = "Feature"
    feature.short_description = "My Feature"
    feature.description = []
    feature.tags = []
    feature.background = mocker.MagicMock(spec=Background)
    feature.background.keyword = "Background"
    feature.background.short_description = "My Background"
    feature.background.steps = []

    # when
    write_feature_header(feature)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            Feature: My Feature
                Background: My Background

            """
        ),
    )


def test_gf_write_feature_header_background_with_steps(disabled_colors, capsys, mocker):
    """
    Test that the Gherkin Formatter properly writes a Feature header
    without Tags and Description but a Background with Steps
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.keyword = "Feature"
    feature.short_description = "My Feature"
    feature.description = []
    feature.tags = []
    feature.background = mocker.MagicMock(spec=Background)
    feature.background.keyword = "Background"
    feature.background.short_description = "My Background"
    first_step = mocker.MagicMock(
        spec=Step,
        keyword="Given",
        used_keyword="Given",
        text="there is a Step",
        doc_string=None,
        data_table=None,
    )
    second_step = mocker.MagicMock(
        spec=Step,
        keyword="When",
        used_keyword="When",
        text="there is a Step",
        doc_string=None,
        data_table=None,
    )
    feature.background.steps = [first_step, second_step]

    # when
    write_feature_header(feature)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            Feature: My Feature
                Background: My Background
                    Given there is a Step
                    When there is a Step
            """
        ),
    )


def test_gf_write_feature_footer_blank_line_if_no_description_and_no_rules(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter writes a blank line after a Feature
    without a description and Rules
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.description = []
    feature.rules = []

    # when
    write_feature_footer(feature)

    # then
    stdout = capsys.readouterr().out
    assert stdout == "\n"


def test_gf_write_feature_footer_no_blank_line_if_description(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter writes no blank line after a Feature
    with a Description
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.description = ["foo"]
    feature.rules = []

    # when
    write_feature_footer(feature)

    # then
    stdout = capsys.readouterr().out
    assert stdout == ""


def test_gf_write_feature_footer_no_blank_line_if_rules(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter writes no blank line after a Feature
    with a Rule
    """
    # given
    feature = mocker.MagicMock(spec=Feature)
    feature.description = []
    feature.rules = ["foo"]

    # when
    write_feature_footer(feature)

    # then
    stdout = capsys.readouterr().out
    assert stdout == ""


def test_gf_write_rule_header(disabled_colors, capsys, mocker):
    """Test that the Gherkin Formatter properly writes a Rule"""
    # given
    rule = mocker.MagicMock(spec=Rule)
    rule.keyword = "Rule"
    rule.short_description = "My Rule"

    # when
    write_rule_header(rule)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            (?P<indentation>    )Rule: My Rule

            """
        ),
    )


def test_gf_write_rule_header_nothing_for_default_rule(disabled_colors, capsys, mocker):
    """Test that the Gherkin Formatter writes no Rule header for a DefaultRule"""
    # given
    rule = mocker.MagicMock(spec=DefaultRule)

    # when
    write_rule_header(rule)

    # then
    stdout = capsys.readouterr().out
    assert stdout == ""


@pytest.mark.parametrize(
    "given_rule_type, expected_indentation",
    [(DefaultRule, " " * 4), (Rule, " " * 8)],
    ids=["DefaultRule", "Rule"],
)
def test_gf_write_scenario_header_without_tags(
    given_rule_type, expected_indentation, disabled_colors, capsys, mocker
):
    """Test that the Gherkin Formatter properly formatter a Scenario Header without Tags"""
    # given
    scenario = mocker.MagicMock(spec=Scenario)
    scenario.keyword = "Scenario"
    scenario.rule = mocker.MagicMock(spec=given_rule_type)
    scenario.short_description = "My Scenario"
    scenario.tags = []

    # when
    write_scenario_header(scenario)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            (?P<indentation>{indentation})Scenario: My Scenario
            """.format(
                indentation=expected_indentation
            )
        ),
    )


@pytest.mark.parametrize(
    "given_rule_type, expected_indentation",
    [(DefaultRule, " " * 4), (Rule, " " * 8)],
    ids=["DefaultRule", "Rule"],
)
def test_gf_write_scenario_header_with_tags(
    given_rule_type, expected_indentation, disabled_colors, capsys, mocker
):
    """Test that the Gherkin Formatter properly formatter a Scenario Header with Tags"""
    # given
    scenario = mocker.MagicMock(spec=Scenario)
    scenario.keyword = "Scenario"
    scenario.rule = mocker.MagicMock(spec=given_rule_type)
    scenario.short_description = "My Scenario"
    first_tag = mocker.MagicMock(spec=Tag)
    first_tag.name = "tag-a"
    second_tag = mocker.MagicMock(spec=Tag)
    second_tag.name = "tag-b"
    scenario.tags = [first_tag, second_tag]

    # when
    write_scenario_header(scenario)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            (?P<indentation>{indentation})@tag-a
            (?P<indentation>{indentation})@tag-b
            (?P<indentation>{indentation})Scenario: My Scenario
            """.format(
                indentation=expected_indentation
            )
        ),
    )


def test_gf_write_scenario_footer_always_a_blank_line(disabled_colors, capsys, mocker):
    """Test that the Gherkin Formatter always writes a blank line after a Scenario"""
    # given
    scenario = mocker.MagicMock(spec=Scenario)

    # when
    write_scenario_footer(scenario)

    # then
    stdout = capsys.readouterr().out
    assert stdout == "\n"


@pytest.mark.parametrize(
    "given_rule_type, expected_indentation",
    [(DefaultRule, " " * 8), (Rule, " " * 12)],
    ids=["DefaultRule", "Rule"],
)
def test_gf_write_step_without_doc_string_without_data_table(
    given_rule_type, expected_indentation, disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly formats a Step without a doc string and data table
    """
    # given
    step = mocker.MagicMock(spec=Step)
    step.keyword = "Given"
    step.used_keyword = "Given"
    step.text = "there is a Step"
    step.doc_string = None
    step.data_table = None
    step.rule = mocker.MagicMock(spec=given_rule_type)

    # when
    write_step(step, step_color_func=lambda x: x)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            (?P<indentation>{indentation})Given there is a Step
            """.format(
                indentation=expected_indentation
            )
        ),
    )


def test_gf_write_step_explicit_indentation_without_doc_string_without_data_table(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly formats a Step with an explicit indentation
    butwithout a doc string and data table
    """
    # given
    step = mocker.MagicMock(spec=Step)
    step.keyword = "Given"
    step.used_keyword = "Given"
    step.text = "there is a Step"
    step.doc_string = None
    step.data_table = None

    # when
    write_step(step, step_color_func=lambda x: x, indentation="   ")

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            (?P<indentation>   )Given there is a Step
            """
        ),
    )


@pytest.mark.parametrize(
    "given_rule_type, expected_indentation",
    [(DefaultRule, " " * 8), (Rule, " " * 12)],
    ids=["DefaultRule", "Rule"],
)
def test_gf_write_step_with_doc_string_without_data_table(
    given_rule_type, expected_indentation, disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly formats a Step with a doc string
    but without a data table
    """
    # given
    step = mocker.MagicMock(spec=Step)
    step.keyword = "Given"
    step.used_keyword = "Given"
    step.text = "there is a Step"
    step.doc_string = """foo
bar
bla
"""
    step.data_table = None
    step.rule = mocker.MagicMock(spec=given_rule_type)

    # when
    write_step(step, step_color_func=lambda x: x)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            (?P<indentation>{indentation})Given there is a Step
            (?P<indentation>{indentation}    )\"\"\"
            (?P<indentation>{indentation}    )foo
            (?P<indentation>{indentation}    )bar
            (?P<indentation>{indentation}    )bla
            (?P<indentation>{indentation}    )\"\"\"
            """.format(
                indentation=expected_indentation
            )
        ),
    )


@pytest.mark.parametrize(
    "given_rule_type, expected_indentation",
    [(DefaultRule, " " * 8), (Rule, " " * 12)],
    ids=["DefaultRule", "Rule"],
)
def test_gf_write_step_with_doc_string_keep_indentation_without_data_table(
    given_rule_type, expected_indentation, disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter properly formats a Step with a doc string
    that has an indentation itself
    but without a data table
    """
    # given
    step = mocker.MagicMock(spec=Step)
    step.keyword = "Given"
    step.used_keyword = "Given"
    step.text = "there is a Step"
    step.doc_string = """foo
    bar
  meh
bla
"""
    step.data_table = None
    step.rule = mocker.MagicMock(spec=given_rule_type)

    # when
    write_step(step, step_color_func=lambda x: x)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            (?P<indentation>{indentation})Given there is a Step
            (?P<indentation>{indentation}    )\"\"\"
            (?P<indentation>{indentation}    )foo
            (?P<indentation>{indentation}    )    bar
            (?P<indentation>{indentation}    )  meh
            (?P<indentation>{indentation}    )bla
            (?P<indentation>{indentation}    )\"\"\"
            """.format(
                indentation=expected_indentation
            )
        ),
    )


@pytest.mark.parametrize(
    "step_state, expected_color",
    [
        pytest.param(State.PASSED, cf.forestGreen, id="State.PASSED => cf.forestGreen"),
        pytest.param(State.FAILED, cf.firebrick, id="State.FAILED => cf.firebrick"),
        pytest.param(State.PENDING, cf.orange, id="State.PENDING => cf.orange"),
        pytest.param(
            State.UNTESTED, cf.deepSkyBlue3, id="State.UNTESTED => cf.deepSkyBlue3"
        ),
    ],
)
def test_gf_write_step_result_without_failure_report(
    step_state, expected_color, world_default_config, disabled_colors, capsys, mocker
):
    """Test that the Gherkin Formatter properly formats a Step result without a Failure Report"""
    # given
    step = mocker.MagicMock(spec=Step)
    step.keyword = "Given"
    step.used_keyword = "Given"
    step.text = "there is a Step"
    step.state = step_state
    step.failure_report = None

    write_step_mock = mocker.patch("radish.formatters.gherkin.write_step")

    # when
    write_step_result(step)

    # then
    write_step_mock.assert_called_once_with(step, expected_color)


def test_gf_write_and_as_keyword_if_not_first_step_of_keyword_context(
    disabled_colors, capsys, mocker
):
    """
    Test that the Gherkin Formatter writes the ``And`` keyword instead of the keyword itself
    if it's not the first Step of this keyword context.
    """
    # given
    first_step = mocker.MagicMock(spec=Step)
    first_step.keyword = "Given"
    first_step.used_keyword = "Given"
    first_step.text = "there is the first Step"
    first_step.doc_string = None
    first_step.data_table = None
    first_step.rule = mocker.MagicMock(spec=DefaultRule)

    second_step = mocker.MagicMock(spec=Step)
    second_step.keyword = "Given"
    second_step.used_keyword = "And"
    second_step.text = "there is the second Step"
    second_step.doc_string = None
    second_step.data_table = None
    second_step.rule = mocker.MagicMock(spec=DefaultRule)

    # when
    write_step(first_step, step_color_func=lambda x: x)
    write_step(second_step, step_color_func=lambda x: x)

    # then
    assert_output(
        capsys,
        dedent_feature_file(
            """
            (?P<indentation>        )Given there is the first Step
            (?P<indentation>        )And there is the second Step
            """
        ),
    )



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""


class Tag:
    """Represents a single Gherkin Tag"""

    def __init__(self, name: str, path: str, line: int) -> None:
        self.name = name
        self.path = path
        self.line = line

    def __repr__(self) -> str:
        return "<Tag: {name} @ {path}:{line}>".format(
            name=self.name, path=self.path, line=self.line
        )  # pragma: no cover



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

from radish.stepregistry import StepImpl, StepRegistry


@pytest.mark.parametrize(
    "step_impl_1, step_impl_2, expected_equal",
    [
        (StepImpl("Given", "pattern", None), StepImpl("Given", "pattern", None), True),
        (StepImpl("Given", "pattern", None), StepImpl("", "pattern", None), False),
        (StepImpl("Given", "pattern", None), StepImpl("Given", "", None), False),
        (
            StepImpl("Given", "pattern", None),
            StepImpl("Given", "pattern", lambda x: x),
            False,
        ),
        (StepImpl("Given", "pattern", None), None, False),
    ],
    ids=[
        "are equal",
        "keyword is not equal",
        "pattern is not equal",
        "func is not equal",
        "right hand side object is no StepImpl object",
    ],
)
def test_step_impl_should_implement_the_equals_protocol(
    step_impl_1, step_impl_2, expected_equal
):
    """Test that StepImpls correctly implement the equals protocol"""
    # when
    are_equal = step_impl_1 == step_impl_2
    hash_equal = hash(step_impl_1) == hash(step_impl_2)

    # then
    assert are_equal == expected_equal
    assert hash_equal == expected_equal


def test_stepregistry_is_initialized_with_no_step_impls():
    """The StepRegistry should be initialized with no registered Step Implementations"""
    # given & when
    registry = StepRegistry()

    # then
    assert registry.step_implementations() == {}


def test_stepregistry_should_allow_to_register_step_impls():
    """The StepRegistry should allow to register a Step Implementation"""
    # given
    registry = StepRegistry()

    # when
    registry.register("Given", "pattern", None)

    # then
    assert registry.step_implementations("Given") == [
        StepImpl("Given", "pattern", None)
    ]


def test_stepregistry_should_gracefully_accept_double_registration():
    """
    The StepRegistry should gracefully accept a duplicate registration of a Step Implementation
    """
    # given
    registry = StepRegistry()
    registry.register("Given", "pattern", None)

    # when
    registry.register("Given", "pattern", None)

    # then
    assert registry.step_implementations("Given") == [
        StepImpl("Given", "pattern", None)
    ]


def test_stepregistry_should_create_one_step_decorator_per_keyword():
    """The StepRegistry should create one Step decorator for each keyword"""
    # given
    registry = StepRegistry()
    context = {}

    # when
    registry.create_step_decorators(context)

    # then
    assert len(context) == 4
    assert "given" in context
    assert "when" in context
    assert "then" in context
    assert "step" in context


@pytest.mark.parametrize("keyword", ["Given", "When", "Then"])
def test_stepregistry_step_decorator_should_register_func_with_proper_keyword(keyword):
    """The StepRegistry should create one Step decorator for each keyword"""
    # given
    registry = StepRegistry()
    context = {}
    registry.create_step_decorators(context)

    # when
    def test_step():
        ...

    test_step = context[keyword.lower()]("pattern")(test_step)

    # then
    assert registry.step_implementations(keyword) == [
        StepImpl(keyword, "pattern", test_step)
    ]


def test_stepregitry_register_func_with_multiple_decorators():
    """The StepRegistry should allow a function to be registered with multiple Step decorators"""
    # given
    registry = StepRegistry()
    context = {}
    registry.create_step_decorators(context)

    # when
    def test_step():
        ...

    test_step = context["given"]("pattern")(test_step)
    test_step = context["when"]("pattern")(test_step)

    # then
    assert registry.step_implementations("Given") == [
        StepImpl("Given", "pattern", test_step)
    ]
    assert registry.step_implementations("When") == [
        StepImpl("When", "pattern", test_step)
    ]


def test_stepregitry_step_decorators_for_all_keywords():
    """The StepRegistry should return the Step Implementations registered
    with the ``step`` decorator for all keywords.
    """
    # given
    registry = StepRegistry()
    context = {}
    registry.create_step_decorators(context)

    # when
    def test_step():
        ...

    test_step = context["step"]("pattern")(test_step)

    # then
    assert registry.step_implementations("Given") == [
        StepImpl("Step", "pattern", test_step)
    ]


def test_stepregistry_module_should_have_global_registry_instance():
    """The radish.stepregistry module should contain a global StepRegistry instance"""
    # given & when
    from radish.stepregistry import registry

    # then
    assert isinstance(registry, StepRegistry)


def test_stepregistry_module_should_have_global_step_decorators():
    """The radish.stepregistry module should contain functions for the Step decorators"""
    # given & when
    from radish.stepregistry import given, when, then, step

    # then
    assert callable(given)
    assert callable(when)
    assert callable(then)
    assert callable(step)


def test_stepregistry_module_level_given_decorator_register_at_global_registry_instance():
    """
    The global module-level given-Step decorators should register
    a Step at the global registry instance
    """
    # given
    from radish.stepregistry import registry, given

    # when
    @given("pattern")
    def given_step():
        ...

    # then
    assert StepImpl("Given", "pattern", given_step) in registry.step_implementations(
        "Given"
    )


def test_stepregistry_module_level_when_decorator_register_at_global_registry_instance():
    """
    The global module-level when-Step decorators should register
    a Step at the global registry instance
    """
    # given
    from radish.stepregistry import registry, when

    # when
    @when("pattern")
    def when_step():
        ...

    # then
    assert StepImpl("When", "pattern", when_step) in registry.step_implementations(
        "When"
    )


def test_stepregistry_module_level_then_decorator_register_at_global_registry_instance():
    """
    The global module-level then-Step decorators should register
    a Step at the global registry instance
    """
    # given
    from radish.stepregistry import registry, then

    # when
    @then("pattern")
    def then_step():
        ...

    # then
    assert StepImpl("Then", "pattern", then_step) in registry.step_implementations(
        "Then"
    )


def test_stepregistry_module_level_step_decorator_register_at_global_registry_instance():
    """
    The global module-level step-Step decorators should register
    a Step at the global registry instance
    """
    # given
    from radish.stepregistry import registry, step

    # when
    @step("pattern")
    def step_step():
        ...

    # then
    assert StepImpl("Step", "pattern", step_step) in registry.step_implementations(
        "Step"
    )



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from pathlib import Path
from typing import List, Dict, Any
from unittest import mock
from collections import OrderedDict, defaultdict
import textwrap

import colorful as cf

import radish.matcher as matcher
import radish.utils as utils
from radish.errors import RadishError, StepImplementationNotFoundError
from radish.models import Step
from radish.stepregistry import StepRegistry


def parse_matcher_config(matcher_config_path):
    """Parse the given matcher config file

    The matcher config file is expected to be a properly
    formatted YAML file, having the following structure:

    .. code-block:: yaml

        - step: <step text>
          should_match | should_not_match: <step implementation func name>
          with_arguments:
            - <args...>


    The YAML file can contain one or more of these ``step`` blocks.
    """
    with matcher_config_path.open("r", encoding="utf-8") as matcher_config_file:
        steps_matcher_config = utils.yaml_ordered_load(matcher_config_file)

    # the matcher config is empty - so we don't have to continue here
    if not steps_matcher_config:
        return None

    # validate the matcher config read from the YAML file
    for step_matcher_config in steps_matcher_config:
        # check if the config really has the `step` key
        if "step" not in step_matcher_config:
            raise RadishError("All Matcher Config must start with the 'step' key")

        # check if the config has either the `should_match` or `should_not_match` key
        if ("should_match" not in step_matcher_config) == (
            "should_not_match" not in step_matcher_config
        ):  # noqa
            raise RadishError(
                "Each Matcher Config must either contain a 'should_match' or 'should_not_match' key"
            )

        # check if the config has any invalid keys
        valid_keys = {"step", "should_match", "should_not_match", "with_arguments"}
        if not set(step_matcher_config.keys()).issubset(valid_keys):
            raise RadishError(
                "All Matcher Config only allow the keys: "
                "{}. The following are not allowed: {}.".format(
                    ", ".join(sorted(valid_keys)),
                    ", ".join(
                        sorted(set(step_matcher_config.keys()).difference(valid_keys))
                    ),
                )
            )

    return steps_matcher_config


def run_matcher_tests(
    matcher_configs: List[Path], coverage_config, step_registry: StepRegistry
):
    """Run the matcher config tests against all Steps in the Registry"""
    #: holds a set of all covered Step Implementations
    #  A Step Implementation only counts as covered if it
    #  was successfully tested against a positive (should_match) test.
    covered_step_impls = set()

    for matcher_config_path in matcher_configs:
        match_config = parse_matcher_config(matcher_config_path)

        # nothing to match against, because config was empty
        if not match_config:
            print(
                cf.orange(
                    "The matcher config {} was empty - Nothing to do :)".format(
                        matcher_config_path
                    )
                )
            )
            continue

        for step_to_match in match_config:
            # create a dummy Step object for the match config
            keyword, text = step_to_match["step"].split(maxsplit=1)
            step = Step(
                step_id=0,
                keyword=keyword,
                used_keyword=keyword,
                text=text,
                doc_string=None,
                data_table=None,
                path=None,
                line=None,
            )
            # provide a Mock Scenario to the Step which
            # is needed for some Step matching functionality.
            step.scenario = mock.MagicMock(name="Mocked Scenario")

            # verify the step matches accordingly to the expectations in the config
            if "should_match" in step_to_match:
                expected_step_func = step_to_match["should_match"]
                if "with_arguments" in step_to_match:
                    expected_step_arguments = step_to_match["with_arguments"]
                else:
                    expected_step_arguments = []  # no Step arguments to match against

                is_matched = assert_step_match(
                    step, expected_step_func, expected_step_arguments, step_registry
                )
                if is_matched:
                    covered_step_impls.add(step.step_impl)
            else:
                assert_step_not_match(
                    step, step_to_match["should_not_match"], step_registry
                )

    # do coverage analysis on the tested Step Implementation coverage
    coverage(covered_step_impls, step_registry, coverage_config)


def assert_step_match(
    step: Step,
    expected_step_func: str,
    expected_step_arguments: List[Dict[str, Any]],
    step_registry: StepRegistry,
):
    """Assert that the Step correctly matches in the Registry"""
    print(
        "{} STEP '{}' SHOULD MATCH {}".format(
            cf.orange(">>"),
            cf.deepSkyBlue3("{} {}".format(step.keyword, step.text)),
            cf.deepSkyBlue3(expected_step_func),
        ),
        end="    ",
        flush=True,
    )

    # match the step text from the config with one from the registry
    try:
        matcher.match_step(step, step_registry)
    except StepImplementationNotFoundError:
        print_failure(None, ["Expected Step Text didn't match any Step Implementation"])
        return False

    # check if Step matches the correct Step Implementation Function
    matched_step_func = step.step_impl.func
    if matched_step_func.__name__ != expected_step_func:
        print_failure(
            matched_step_func,
            [
                "Expected Step Text matched {} instead of {}".format(
                    matched_step_func.__name__, expected_step_func
                )
            ],
        )
        return False

    # check if the Step has a match with the correct arguments
    if expected_step_arguments:
        # merge the Step's keyword and positional arguments into one dict
        args, kwargs = step.step_impl_match.evaluate()
        actual_step_arguments = utils.get_func_pos_args_as_kwargs(
            matched_step_func, args
        )
        actual_step_arguments.update(kwargs)

        # turn the list of single-item-dicts to a multi-item dict
        # -> [{1: 2}, {3: 4}] --> {1: 2, 3: 4}
        # NOTE(TF) for Python 3.5 test reproducibility we need an OrderedDict -.^
        expected_step_arguments = OrderedDict(
            (
                argpair
                for argpairs in expected_step_arguments
                for argpair in argpairs.items()
            )
        )
        errors = assert_step_arguments(actual_step_arguments, expected_step_arguments)

        if errors:
            print_failure(matched_step_func, errors)
            return False

    print(cf.bold_forestGreen("✔"))
    return True


def assert_step_not_match(
    step: Step, expected_step_func: str, step_registry: StepRegistry
):
    """Assert that the Step doesn't match a Step Implementation from the Registry"""
    print(
        "{} STEP '{}' SHOULD NOT MATCH {}".format(
            cf.orange(">>"),
            cf.deepSkyBlue3("{} {}".format(step.keyword, step.text)),
            cf.deepSkyBlue3(expected_step_func if expected_step_func else "ANY"),
        ),
        end="    ",
        flush=True,
    )

    # match the step text from the config with one from the registry
    try:
        matcher.match_step(step, step_registry)
    except StepImplementationNotFoundError:
        print(cf.bold_forestGreen("✔"))
        return True

    matched_step_func = step.step_impl.func
    if matched_step_func.__name__ == expected_step_func:
        print_failure(
            matched_step_func,
            [
                "Expected Step Text matched {} but it shouldn't".format(
                    expected_step_func
                )
            ],
        )
        return False

    print(cf.bold_forestGreen("✔"))
    return True


def assert_step_arguments(actual_step_arguments, expected_step_arguments):
    """Test the matches between the actual Step arguments and the expected step arguments"""
    # collect all argument errors and return them all at once
    errors = []

    for arg_name, arg_value in expected_step_arguments.items():
        # if an expected Step argument isn't present in the actual Step arguments
        if arg_name not in actual_step_arguments:
            errors.append(
                'Expected argument "{0}" is not in matched arguments {1}'.format(
                    arg_name, sorted(set(actual_step_arguments.keys()))
                )
            )
            continue

        #: holds a flag to indicate that the types representation
        #  should be used for a match and not the actual object value.
        use_repr = False

        # check if argument value is a dict, if yes we'll do thorough comparison
        if isinstance(arg_value, dict) and "type" not in arg_value.keys():
            # in case the argument should be treated as an actual Python dict
            _type = "dict"
            value = arg_value
        elif isinstance(arg_value, dict):
            _type = arg_value["type"]
            value = arg_value["value"]
            # Use repr protocol to match argument values
            use_repr = "use_repr" in arg_value and arg_value["use_repr"]

            # check if value should be casted to the given type
            if "cast" in arg_value and arg_value["cast"] is True:
                obj_type = utils.locate_python_object(_type)
                if obj_type is None:
                    errors.append(
                        'Cannot cast to type "{0}" because it is unknown'.format(_type)
                    )
                    continue

                try:
                    value = obj_type(value)
                except Exception:
                    errors.append(
                        'Failed to cast "{0}" to given type "{1}"'.format(value, _type)
                    )
                    continue
        else:
            _type = type(arg_value).__name__
            value = arg_value

        if not use_repr and _type != type(value).__name__:
            errors.append(
                "Conflicting argument configuration: given value is actually of "
                'type "{0}" although it should match a value of type "{1}"'.format(
                    type(value).__name__, _type
                )
            )
            continue

        if type(actual_step_arguments[arg_name]).__name__ != _type:
            errors.append(
                'Expected argument "{0}" is of type "{2}" but should be "{1}"'.format(
                    arg_name, type(actual_step_arguments[arg_name]).__name__, _type
                )
            )
            continue

        matched = None
        if use_repr:
            matched = repr(actual_step_arguments[arg_name]) != value
        else:
            matched = actual_step_arguments[arg_name] != value

        if matched:
            errors.append(
                'Expected argument "{0}" with value "{1}" does not match value "{2}"'.format(
                    arg_name, value, actual_step_arguments[arg_name]
                )
            )
    return errors


def print_failure(matched_step_func, errors: List[str]):
    """Print a failure message for matching errors"""
    print(cf.bold_firebrick("✘"), end=" ", flush=True)

    if matched_step_func is not None:
        func_code = matched_step_func.__code__
        print(
            cf.firebrick(
                "(at {}:{})".format(func_code.co_filename, func_code.co_firstlineno)
            ),
            end="",
            flush=True,
        )

    print(flush=True)

    for error in errors:
        print(cf.firebrick("  - {}".format(error)), flush=True)


def coverage(covered_step_impls, step_registry, coverage_config):
    """Analyse the coverage of the registered Step Implementations"""
    # get all Step Implementations
    all_step_impls = {
        step_impl
        for step_impls in step_registry.step_implementations().values()
        for step_impl in step_impls
    }

    if coverage_config.show_missing or coverage_config.show_missing_templates:
        # get all Step Implementations not covered
        not_covered_step_impls = list(all_step_impls.difference(covered_step_impls))

        if not not_covered_step_impls:
            print("Everything is covered!")
            return

        # sort not covered steps according their line number
        not_covered_step_impls.sort(key=lambda x: x.func.__code__.co_firstlineno)

        # order missing Step Implementations by the module they are implemented in
        ordered_not_covered_step_impls = defaultdict(list)
        for not_covered_step_impl in not_covered_step_impls:
            step_module = not_covered_step_impl.func.__code__.co_filename
            ordered_not_covered_step_impls[step_module].append(not_covered_step_impl)

        # output not covered Step Implementations per module
        for module_path, step_impls in ordered_not_covered_step_impls.items():
            print("Missing from: {}".format(cf.bold(module_path)))
            for step_impl in step_impls:
                print(
                    "  - {}:{}".format(
                        step_impl.func.__name__, step_impl.func.__code__.co_firstlineno
                    )
                )

        if coverage_config.show_missing_templates:
            # output template YAML to test the not covered Step Implementations
            print()
            print(
                "Add the following to your matcher-config.yml "
                "to cover the missing Step Implementations:"
            )
            for step_impl in not_covered_step_impls:
                print(
                    textwrap.dedent(
                        """
                      {cf.gray}# testing Step Implementation at {path}:{lineno}{cf.reset}
                      {cf.forestGreen}-{cf.reset} {cf.bold_dodgerBlue}step{cf.reset}{cf.red}: {cf.italic_lightSkyBlue}"<insert sample Step Text here>"{cf.reset}
                        {cf.bold_dodgerBlue}should_match{cf.reset}{cf.red}:{cf.reset} {step_impl_func_name}
                    """.format(  # noqa
                            cf=cf,
                            path=step_impl.func.__code__.co_filename,
                            lineno=step_impl.func.__code__.co_firstlineno,
                            step_impl_func_name=step_impl.func.__name__,
                        )
                    ).rstrip()
                )

                # check if function has arguments which can be tested
                arg_names = step_impl.func.__code__.co_varnames[
                    1 : step_impl.func.__code__.co_argcount
                ]  # without the `step` argument
                if arg_names:
                    argument_match_lines = [
                        "    {cf.forestGreen}-{cf.reset} {cf.bold_dodgerBlue}{arg}{cf.reset}{cf.red}:{cf.reset} {cf.italic}<insert argument value here>{cf.reset}".format(  # noqa
                            cf=cf, arg=x
                        )
                        for x in arg_names
                    ]
                    print(
                        "  {cf.bold_dodgerBlue}with_arguments{cf.reset}{cf.red}:{cf.reset}".format(
                            cf=cf
                        )
                    )  # noqa
                    print("\n".join(argument_match_lines))
                print()



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.parsetyperegistry import custom_type


@custom_type("int", r"[-+]?[0-9]+")
def int_type(text):
    """Custom Parse Type to parse an integer value.

    The integer value must be of the format: [-+]?[0-9]+
    """
    return int(text)


@custom_type(
    "float", r"[-+]? (?: (?: \d* \. \d+ ) | (?: \d+ \.? ) )(?: [Ee] [+-]? \d+ ) ?"
)
def float_type(text):
    r"""Custom Parse Type to parse a float value.

    The float value must be of the format:
        [-+]? (?: (?: \d* \. \d+ ) | (?: \d+ \.? ) )(?: [Ee] [+-]? \d+ ) ?
    """
    return float(text)


@custom_type("word", r"\S+")
def word_type(text):
    r"""Custom Parse Type to parse a word.

    The word value must be of the format: \S+
    """
    return str(text)


@custom_type(
    "bool",
    r"(0|1|yes|Yes|YES|y|Y|no|No|NO|n|N|true|True|TRUE|false|False|FALSE|on|On|ON|off|Off|OFF)",
)
def boolean_type(text):
    """Custom Parse Type to parse a boolean value.

    The same values are parsed as YAML does:
        http://yaml.org/type/bool.html

    Plus 0 and 1
    """
    text = text.lower()
    return text == "1" or text.startswith("y") or text == "true" or text == "on"


@custom_type("QuotedString", r'"(?:[^"\\]|\\.)*"')
def quoted_string_type(text):
    """Custom Parse Type to parse a quoted string.

    Double quotes (") have to be escaped with a
    backslash within the double quotes.
    """
    return text[1:-1]


@custom_type("MathExpression", r"[0-9 +\-*/%.e]+")
def math_expression_type(text):
    """Custom Parse Type which expects a valid mathematical expression

    :param str text: the text which was matched as math expression

    :returns: calculated float number from the math expression
    :rtype: float
    """
    return float(eval(text))



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

from radish.models.state import State
from radish.models.background import Background


@pytest.mark.parametrize(
    "given_step_states, expected_state",
    [
        ([State.PASSED, State.PASSED, State.PASSED], State.PASSED),
        ([State.PASSED, State.UNTESTED, State.PASSED], State.UNTESTED),
        ([State.PASSED, State.SKIPPED, State.UNTESTED], State.SKIPPED),
        ([State.PASSED, State.PENDING, State.SKIPPED], State.PENDING),
        ([State.PASSED, State.FAILED, State.PENDING], State.FAILED),
        ([State.PASSED, State.RUNNING, State.FAILED], State.RUNNING),
    ],
    ids=[
        "[State.PASSED, State.PASSED, State.PASSED] -> State.PASSED",
        "[State.PASSED, State.UNTESTED, State.PASSED] -> State.UNTESTED",
        "[State.PASSED, State.SKIPPED, State.UNTESTED] -> State.SKIPPED",
        "[State.PASSED, State.PENDING, State.SKIPPED] -> State.PENDING",
        "[State.PASSED, State.FAILED, State.PENDING] -> State.FAILED",
        "[State.PASSED, State.RUNNING, State.FAILED] -> State.RUNNING",
    ],
)
def test_background_should_correct_state(given_step_states, expected_state, mocker):
    """A Background should return the worst State of its Steps State as its own State"""
    # given
    background = Background(
        "Background",
        "Background",
        None,
        None,
        [mocker.MagicMock(state=s) for s in given_step_states],
    )

    # when
    actual_state = background.state

    # then
    assert actual_state == expected_state



"""
radish
~~~~~~

the root from red to green.  BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.models.background import Background  # noqa
from radish.models.constant_tag import ConstantTag  # noqa
from radish.models.feature import Feature  # noqa
from radish.models.precondition_tag import PreconditionTag  # noqa
from radish.models.rule import DefaultRule, Rule  # noqa
from radish.models.scenario import Scenario  # noqa
from radish.models.scenario_loop import ScenarioLoop  # noqa
from radish.models.scenario_outline import ScenarioOutline  # noqa
from radish.models.state import State  # noqa
from radish.models.step import Step  # noqa
from radish.models.tag import Tag  # noqa



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import sys
from pathlib import Path

import click
import colorful as cf

from radish.config import Config
from radish.terrain import world
from radish.parser.core import FeatureFileParser
from radish.errors import RadishError
from radish.logger import enable_radish_debug_mode_click_hook, logger


def expand_feature_files(ctx, param, feature_files):
    """Expand the given feature files

    Expanding directories recursively for Feature Files
    """
    expanded_feature_files = []
    for feature_file_location in (Path(f) for f in feature_files):
        if feature_file_location.is_dir():
            expanded_feature_files.extend(
                list(feature_file_location.glob("**/*.feature"))
            )
        else:
            expanded_feature_files.append(feature_file_location)

    return expanded_feature_files


@click.command()
@click.version_option()
@click.help_option("--help", "-h")
@click.option(
    "--debug",
    "-d",
    "enable_debug_mode",
    is_flag=True,
    help="Enable debug mode for radish itself",
    callback=enable_radish_debug_mode_click_hook,
)
@click.option(
    "--no-ansi",
    "no_ansi",
    is_flag=True,
    help=(
        "Turn off all ANSI sequences (colors, line rewrites) ."
        "This option is mainly used by the formatters"
    ),
)
@click.option(
    "--resolve-preconditions",
    "resolve_preconditions",
    is_flag=True,
    help="Resolve @preconditions when parsing the Feature Files",
)
@click.argument(
    "feature_files",
    nargs=-1,
    type=click.Path(exists=True),
    callback=expand_feature_files,
)
def cli(**kwargs):
    """radish - The root from red to green. BDD tooling for Python.

    Parse and Pretty Print the raw radish AST for the given Feature Files

    Provide the Feature Files to run in FEATURE_FILES.
    """
    config = Config(kwargs)
    world.config = config

    # turn of ANSI colors if requested
    if config.no_ansi:
        cf.disable()

    parser = FeatureFileParser(
        ast_transformer=None, resolve_preconditions=config.resolve_preconditions
    )

    for feature_file in config.feature_files:
        logger.info("Parsing Feature File %s", feature_file)
        try:
            feature_ast = parser.parse(feature_file)
            if feature_ast:
                print(feature_ast.pretty())
        except RadishError as exc:
            print("", flush=True)
            print(
                "An error occured while parsing the Feature File {}:".format(
                    feature_file
                ),
                flush=True,
            )
            print(exc, flush=True)
            sys.exit(1)


if __name__ == "__main__":
    cli()
